import javax.sound.midi.*;
import java.io.*;
import javax.swing.*;
import java.util.*;

/**
 * Moosique - The MIDI Tracker
 * 
 * Main class that handles initiation, IO and sound.
 * 
 * @author  Einar Pehrson
 */
 
public class Moosique {

	private static MooGUI gui;
	private static Sequence seq;
	private static Sequencer sequencer;
	private static Synthesizer synthesizer;
	private static MidiChannel[] channels;
	private static MidiChannel activeChannel;
	private static MidiEvent[] timeSignatures, tempoChanges;
	private static ArrayList emptyTracks;

	private static String filename, fileArg;
	private static long editPosition;
	private static boolean makeGUI = true, isEdited = false, drawEmptyTracks = false;
	private static Thread player;
	public static final int DEFAULT_RESOLUTION = 96, DEFAULT_TRACKS = 4;

	/** 
	 * Starts the application.
	 */
	public static void main (String[] args) {
		System.out.println("\nMoosique version 1.0\n");

		// Parses command-line arguments.
		for (int i = 0; i < args.length; i++) {
			if (args[i].equals("-n")) {makeGUI = false;}
			else if (fileArg == null) {fileArg = args[i];}
		}

		// Acquires MIDI devices and connects them.
		System.out.print("Initializing MIDI devices.");
		try {
			sequencer = MidiSystem.getSequencer();
			System.out.print(".");
			sequencer.open();
			synthesizer = MidiSystem.getSynthesizer();
			System.out.print(".");
			synthesizer.open();
			sequencer.getTransmitter().setReceiver(synthesizer.getReceiver());
			channels = synthesizer.getChannels();
			setActiveChannel(0);
		} catch (MidiUnavailableException e) {
			System.out.println("Failed, quitting.");
//			System.exit(1);
		}
		System.out.println("Done");

		//If a filename is given as the command-line argument, attempts to load a sequence from the file.
		if (fileArg != null) {
			System.out.print("Loading MIDI sequence from " + fileArg + "...");
			if (!load(fileArg)) {
				System.out.println("Failed");
				clearSequence();
			} else {
				System.out.println("Done");
			}
		} else {
			// Otherwise creates a new empty one.
			clearSequence();
		}

		// Builds GUI, unless n-flag is set.
		if (makeGUI) {
			System.out.print("Building GUI...");
			try {
				UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
			} catch (Exception e) {}
			gui = new MooGUI(seq);
			System.out.println("Done");
		} else {
			System.out.print("Playing...");
			play();
			while (sequencer.isRunning()) {}
			System.out.println("Done");
			quit();
		}
	}

	/** 
	 * Returns the GUI.
	 * @return the GUI
	 */
	public static MooGUI getGUI() {
		return gui;
	}

	/** 
	 * Returns the current sequence.
	 * @return the current sequence
	 */
	public static Sequence getSequence() {
		return seq;
	}

	/** 
	 * Returns the current sequencer.
	 * @return the current sequencer
	 */
	public static Sequencer getSequencer() {
		return sequencer;
	}

	/** 
	 * Returns the MidiChannels of the selected synthesizer.
	 * @return the available MidiChannels
	 */
	public static MidiChannel[] getChannels() {
		return channels;
	}

	/** 
	 * Returns the MidiChannels of the selected synthesizer.
	 * @return the available MidiChannels
	 */
	public static MidiChannel getChannel(int i) {
		return channels[i];
	}

	/** 
	 * Returns the currently active MidiChannel.
	 * @return the active MidiChannel
	 */
	public static MidiChannel getActiveChannel() {
		return activeChannel;
	}

	/** 
	 * Sets the currently active MidiChannel.
	 * @param channel	the number of the MidiChannel to activate
	 */
	public static void setActiveChannel(int channel) {
		activeChannel = channels[channel];
	}

	/** 
	 * Replaces the current sequence with a new one, holding three empty tracks.
	 */
	public static void clearSequence() {
		// Creates a new sequence and sends it to the sequencer.
		try {
			seq = new Sequence(Sequence.PPQ, DEFAULT_RESOLUTION, DEFAULT_TRACKS);
			sequencer.setSequence(seq);
		} catch (InvalidMidiDataException e) {}
		// Sends sequence to GUI.
		if (gui != null) gui.setSequence(seq);
	}

	/** 
	 * Starts playback of the current sequence.
	 */
	public static void play() {
		sequencer.setTickPosition(editPosition);
		resume();
	}

	/** 
	 * Pauses playback of the current sequence.
	 */
	public static void pause() {
		if (sequencer.isRunning()) {
			sequencer.stop();
		}
		if (player != null) player.interrupt();
	}

	/** 
	 * Resumes playback of the current sequence.
	 */
	public static void resume() {
		gui.update(0);
		sequencer.start();

		// Disables input to volatile components
		// gui.disable();

		// Creates the visualisation thread and starts it.
		player = new Thread () {
			public void run() {
				long ticksPerSixteenth = seq.getResolution()/4;
				System.out.println("Ticks/16: " + ticksPerSixteenth);
				long position = sequencer.getTickPosition();
				while(sequencer.isRunning()) {
					long pos = sequencer.getTickPosition();
					long tickDiff = pos - position;
					System.out.print(" ... " + tickDiff);
					position = pos;

					// Updates the GUI with the current tick position.
					gui.update(sequencer.getTickPosition());

					// Puts the thread to sleep for as long as it takes
					// the sequencer to reach the next sixteenth.
					try {
						sleep((long)((15000 / getTempo()) * (tickDiff / ticksPerSixteenth)));
					} catch (InterruptedException e) {
						Moosique.stop();
					}
				}
				Moosique.stop();
			}
		};
		player.start();
	}

	/** 
	 * Stops playback of the current sequence.
	 */
	public static void stop() {
		if (sequencer.isRunning()) {
			sequencer.stop();
		}
		sequencer.setTickPosition(editPosition);
		if (player != null) player.interrupt();
		gui.update((long)0);
	}

	/** 
	 * Returns the current editing position of the sequencer.
	 * @return the tick position
	 */
	public static long getEditPosition() {
		return editPosition;
	}

	/** 
	 * Sets the current editing position of the sequencer.
	 * @param ticks		the tick position
	 */
	public static void setEditPosition(long ticks) {
		editPosition = ticks;
	}

	/** 
	 * Returns the tempo of the current sequence.
	 * @return the tick position
	 */
	public static int getTempo() {
		return 120;
		// if (tempoMsg == null) return 0;
	}

	/** 
	 * Sets the current editing position of the sequencer.
	 * @param ticks		the tick position
	 */
	public static void setTempo(int bpm) {
		// tempoMsg
	}

	/** 
	 * Returns the tempo of the current sequence.
	 * @return the tick position
	 */
	public static int[] getTimeSig() {
		int[] ts = {4, 4};
		return ts;
		// if (timeSigMsg == null) return 0;
	}

	/** 
	 * Sets the current editing position of the sequencer.
	 * @param ticks		the tick position
	 */
	public static void setTimeSig(int bpm) {
		// timeSigMsg
	}

	/** 
	 * Returns true if the current sequence has been edited.
	 * @return the tick position
	 */
	public static boolean isEdited() {
		return isEdited;
	}

	/** 
	 * Sets the current sequence as edited, which implies prompts when loading a new sequence.
	 */
	public static void setEdited() {
		isEdited = true;
	}

	/** 
	 * Rewinds the current sequence the given number of measures.
	 * @param measures	the number of measures to rewind
	 */
	public static void rewind(long ticks) {
		editPosition -= ticks;
	}

	/** 
	 * Fast forwards the current sequence the given number of measures.
	 * @param measures	the number of measures to fast forward
	 */
	public static void forward(long ticks) {
		editPosition += ticks;
	}

	/** 
	 * Shows the given message in the status bar.
	 * @param text	the message to show
	 */
	public static boolean shouldBeDrawn(Track track) {
		if (drawEmptyTracks) return true;
		else return (!emptyTracks.contains(track));
	}


	/** 
	 * Shows the given message in the status bar.
	 * @param text	the message to show
	 */
	public static void setDrawEmptyTracks(boolean state) {
		drawEmptyTracks = state;
	}

	/** 
	 * Loads the MooSequence in the given file.
	 * @param filename	the filename to use
	 */
	public static boolean load(String file) {
		// Loads sequence from file
		filename = file;
		try {
			seq = MidiSystem.getSequence(new File(filename));
		} catch (InvalidMidiDataException e) {
			return false;
		} catch (IOException e) {
			return false;
		}
		isEdited = false;

		Track[] tracks = seq.getTracks();
		emptyTracks = new ArrayList();

		// Searches track 0 for changes in tempo and time signature.
		MidiEvent event;
		MetaMessage metaMsg;
		ArrayList ts = new ArrayList(), tc = new ArrayList();
		for (int i = 0; i < tracks[0].size(); i++) {
			event = tracks[0].get(i);
			if (event.getMessage().getStatus() == MetaMessage.META) {
				metaMsg = (MetaMessage)event.getMessage();
				switch(metaMsg.getType()) {
					case 81: tc.add(event); break;
					case 88: ts.add(event);
				}
			}
		}
//		timeSignatures = ts.toArray(timeSignatures);
//		tempoChanges = tc.toArray(tempoChanges);

		// Wraps each NoteOn event with its NoteOff event in a MooNote
		ArrayList noteOns, noteOffs;
		for (int i = 0; i < tracks.length; i++) {
			// Searches the sequence for NoteOn and NoteOff events
			noteOns = new ArrayList(tracks[i].size() / 2);
			noteOffs = new ArrayList(tracks[i].size() / 2);
			for (int j = 0; j < tracks[i].size(); j++) {
				event = tracks[i].get(j);
				if (event.getMessage().getStatus() >= 144 &&
				    event.getMessage().getStatus() < 160) noteOns.add(event);
				if (event.getMessage().getStatus() >= 128 &&
				    event.getMessage().getStatus() < 144) noteOffs.add(event);
			}
			noteOns.trimToSize();
			noteOffs.trimToSize();
			boolean isEmpty = (noteOns.size() == 0);
			String text = "Track " + i + " has " + noteOns.size() + "/" + noteOffs.size() + "/" + tracks[i].size();
			if (isEmpty) {
				text += " and will be removed.";
				emptyTracks.add(tracks[i]);
			}
			System.out.println(text);
			
			// Sorts the note lists by tick position.
			Comparator c = new Comparator() {
				public int compare(Object o1, Object o2) {
					return (int)(((MidiEvent)o1).getTick() - ((MidiEvent)o2).getTick());
				}
			};
			Collections.sort(noteOns, c);
			Collections.sort(noteOffs, c);

			// For each NoteOn event, finds its NoteOff event and replaces it with a MooNote.
			Iterator iOn = noteOns.iterator(), iOff;
			MidiEvent on, off = null, nextOff;
			ShortMessage onMsg, nextOffMsg;
			while(iOn.hasNext()) {
				on = (MidiEvent)iOn.next();
				onMsg = (ShortMessage)on.getMessage();
				iOff = noteOffs.iterator();
				while(iOff.hasNext()) {
					nextOff = (MidiEvent)iOff.next();
					nextOffMsg = (ShortMessage)nextOff.getMessage();
					if(onMsg.getChannel() == nextOffMsg.getChannel() &&
					   onMsg.getData1() == nextOffMsg.getData1() &&
					   c.compare(nextOff, on) > 0) {
						off = nextOff;
						iOff.remove();
						break;
					}
					   	
				}
				tracks[i].remove(on);
				if (off != null) {
					tracks[i].add(new MooNote(on, off));
				} else {
					tracks[i].add(new MooNote(on));
				}
				iOn.remove();
			}
		}

	/*
		Collections.sort(track[i].events, new Comparator() {
			public int compare(Object o1, Object o2) {
				return ((MidiEvent)o2).getTick() - ((MidiEvent)o1).getTick();
			}
	        });

		// Searches the sequence for NoteOn events
		MidiEvent noteOn, noteOff = null, nextEvent;
		MidiMessage nextMsg;
		ShortMessage shortMsg;

		for (int i = 0; i < tracks.length; i++) {
			for (int j = 0; j < tracks[i].size(); j++) {
				noteOn = tracks[i].get(j);
				if (noteOn.getMessage().getStatus() == ShortMessage.NOTE_ON) {
					// Finds the corresponding NoteOff event
					for (int k = j + 1; k < tracks[i].size(); k++) {
						nextEvent = tracks[i].get(k);
						nextMsg = nextEvent.getMessage();
						if (nextMsg instanceof ShortMessage) {
							shortMsg = (ShortMessage) nextMsg;
							if (shortMsg.getCommand() == ShortMessage.NOTE_OFF && shortMsg.getChannel() == ((ShortMessage)noteOn.getMessage()).getChannel() && shortMsg.getData1() == ((ShortMessage)noteOn.getMessage()).getData1()) {
								noteOff = nextEvent;
								break;
							}
						}
					}
					// Replaces the NoteOn event with a MooNote, if possible with the corresponding NoteOff event
					tracks[i].remove(noteOn);
					if (noteOff != null) {
						tracks[i].add(new MooNote(noteOn, noteOff));
					} else {
						tracks[i].add(new MooNote(noteOn));
					}
				}
			}
		}
*/
		// Sends sequence to GUI and sequencer, then returns
		if (gui != null) gui.setSequence(seq);
		try {
			sequencer.setSequence(seq);
		} catch (InvalidMidiDataException e) {}
		return true;
	}

	/** 
	 * Saves the current sequence to the given filename
	 * @param file	the filename to use
	 */
	public static void saveAs(String file) {
		try {
			MidiSystem.write(seq, 1, new File(filename));
		} catch (IOException e) {}
		filename = file;
		gui.setStatus("Saved " + file);
	}

	/** 
	 * Saves the current sequence to the previously given filename.
	 */
	public static void save() {
		saveAs(filename);
	}

	/** 
	 * Releases all reserved devices and exits the program.
	 */
	public static void quit() {
		if (sequencer.isOpen()) sequencer.close();
		if (synthesizer.isOpen()) synthesizer.close();
		System.exit(0);
	}
}